from typing import List

from pygitguardian.iac_models import IaCFileResult, IaCVulnerability


def filter_unignored_incidents(
    incidents: List[IaCVulnerability],
) -> List[IaCVulnerability]:
    """Removes ignored incidents from the given list"""
    return [incident for incident in incidents if incident.status != "IGNORED"]


def filter_unignored_files(files: List[IaCFileResult]) -> List[IaCFileResult]:
    unignored_files: List[IaCFileResult] = []
    for file in files:
        unignored_incidents = filter_unignored_incidents(file.incidents)
        if len(unignored_incidents) > 0:
            unignored_files.append(
                IaCFileResult(filename=file.filename, incidents=unignored_incidents)
            )
    return unignored_files



from dataclasses import fields, is_dataclass
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Union, overload

import yaml
import yaml.parser
import yaml.scanner

from ggshield.core.constants import (
    AUTH_CONFIG_FILENAME,
    DEFAULT_CONFIG_FILENAME,
    USER_CONFIG_FILENAMES,
)
from ggshield.core.dirs import get_config_dir, get_user_home_dir
from ggshield.core.errors import UnexpectedError


def replace_in_keys(data: Union[List, Dict], old_char: str, new_char: str) -> None:
    """Replace old_char with new_char in data keys."""
    if isinstance(data, dict):
        for key, value in list(data.items()):
            replace_in_keys(value, old_char=old_char, new_char=new_char)
            if old_char in key:
                new_key = key.replace(old_char, new_char)
                data[new_key] = data.pop(key)
    elif isinstance(data, list):
        for element in data:
            replace_in_keys(element, old_char=old_char, new_char=new_char)


def load_yaml_dict(path: Union[str, Path]) -> Optional[Dict[str, Any]]:
    path = Path(path)
    if not path.exists():
        return None

    with path.open() as f:
        try:
            data = yaml.safe_load(f) or {}
        except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
            message = f"{path} is not a valid YAML file:\n{str(e)}"
            raise ValueError(message)

    if not isinstance(data, dict):
        raise ValueError(f"{path} should be a dictionary.")

    return data


def save_yaml_dict(data: Dict[str, Any], path: Union[str, Path]) -> None:
    p = Path(path)
    p.parent.mkdir(parents=True, exist_ok=True)
    with p.open("w") as f:
        try:
            stream = yaml.dump(data, indent=2, default_flow_style=False)
            f.write(stream)
        except Exception as e:
            raise UnexpectedError(f"Failed to save config to {path}:\n{str(e)}") from e


def get_auth_config_filepath() -> Path:
    return get_config_dir() / AUTH_CONFIG_FILENAME


def get_global_path(filename: str) -> Path:
    return get_user_home_dir() / filename


@overload
def find_global_config_path(*, to_write: Literal[False] = False) -> Optional[Path]:
    ...


@overload
def find_global_config_path(*, to_write: Literal[True]) -> Path:
    ...


def find_global_config_path(*, to_write: bool = False) -> Optional[Path]:
    """
    Returns the path to the user global config file (the file in the user home
    directory).
    If there is no such file:
    - If `to_write` is False, returns None.
    - If `to_write` is True, returns the path to the default file.

    This means the function never returns None if `to_write` is True.
    """
    for filename in USER_CONFIG_FILENAMES:
        path = get_global_path(filename)
        if path.exists():
            return path
    return get_global_path(DEFAULT_CONFIG_FILENAME) if to_write else None


def find_local_config_path() -> Optional[Path]:
    for filename in USER_CONFIG_FILENAMES:
        path = Path(filename)
        if path.exists():
            return path
    return None


def update_from_other_instance(dst: Any, src: Any) -> None:
    """
    Update `dst` with fields from `src` if they are set.
    `src` must be the same class or a subclass of `dst`.
    """
    assert isinstance(src, dst.__class__)
    for field_ in fields(src):
        name = field_.name
        value = src.__dict__[name]
        if value is None:
            continue
        if isinstance(value, list):
            dst.__dict__[name].extend(value)
        elif isinstance(value, set):
            dst.__dict__[name].update(value)
        elif is_dataclass(value):
            update_from_other_instance(dst.__dict__[name], value)
        else:
            dst.__dict__[name] = value


def remove_common_dict_items(dct: Dict, reference_dct: Dict) -> Dict:
    """
    Returns a copy of `dct` with all items already in `reference_dct` removed.
    """

    result_dct = dict()
    for key, value in dct.items():
        reference_value = reference_dct[key]

        if isinstance(value, dict):
            value = remove_common_dict_items(value, reference_value)
            # Remove empty dicts
            if not value:
                continue
        else:
            if value == reference_value:
                continue

        result_dct[key] = value

    return result_dct


def remove_url_trailing_slash(url: str) -> str:
    if url[-1] == "/":
        return url[:-1]
    else:
        return url



from typing import Any, Dict, cast

from ggshield.verticals.iac.collection.iac_diff_scan_collection import (
    IaCDiffScanCollection,
)
from ggshield.verticals.iac.collection.iac_path_scan_collection import (
    IaCPathScanCollection,
)
from ggshield.verticals.iac.output.iac_output_handler import IaCOutputHandler
from ggshield.verticals.iac.output.schemas import (
    IaCJSONScanDiffResultSchema,
    IaCJSONScanResultSchema,
)


class IaCJSONOutputHandler(IaCOutputHandler):
    def _process_scan_impl(self, scan: IaCPathScanCollection) -> str:
        scan_dict = IaCJSONOutputHandler.create_scan_dict(scan)
        text = IaCJSONScanResultSchema().dumps(scan_dict)
        return cast(str, text)

    def _process_skip_scan_impl(self) -> str:
        return "{}"

    def _process_skip_diff_scan_impl(self) -> str:
        return "{}"

    def _process_diff_scan_impl(self, scan: IaCDiffScanCollection) -> str:
        scan_dict = IaCJSONOutputHandler.create_diff_scan_dict(scan)
        text = IaCJSONScanDiffResultSchema().dumps(scan_dict)
        return cast(str, text)

    @staticmethod
    def create_scan_dict(scan: IaCPathScanCollection) -> Dict[str, Any]:
        result_without_ignored = scan.get_result_without_ignored()
        if result_without_ignored is None:
            return {
                "id": scan.id,
                "type": scan.type.value,
                "total_incidents": 0,
                "entities_with_incidents": [],
            }
        scan_dict = result_without_ignored.to_dict()
        scan_dict["total_incidents"] = 0

        for entity in scan_dict["entities_with_incidents"]:
            total_incidents = len(entity["incidents"])
            entity["total_incidents"] = total_incidents
            scan_dict["total_incidents"] += total_incidents

        return scan_dict

    @staticmethod
    def create_diff_scan_dict(scan: IaCDiffScanCollection) -> Dict[str, Any]:
        result_without_ignored = scan.get_result_without_ignored()
        if result_without_ignored is None:
            return {
                "id": scan.id,
                "type": scan.type.value,
                "total_incidents": 0,
                "entities_with_incidents": {
                    "unchanged": [],
                    "new": [],
                    "deleted": [],
                },
            }
        scan_dict = result_without_ignored.to_dict()

        for category in scan_dict["entities_with_incidents"]:
            for file_result in scan_dict["entities_with_incidents"][category]:
                file_result["total_incidents"] = len(file_result["incidents"])

        return scan_dict



import sys
from typing import Any

import rich.markup
from rich.console import Console
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn
from typing_extensions import Self

from ggshield.core.ui.scanner_ui import ScannerUI

from ..ggshield_ui import GGShieldProgress, GGShieldUI
from .rich_scanner_ui import RichMessageOnlyScannerUI, RichProgressScannerUI


class RichGGShieldProgress(GGShieldProgress):
    def __init__(self, ui: "RichGGShieldUI", total: int) -> None:
        super().__init__(ui)

        self.progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("{task.completed} / {task.total}"),
            console=ui.console,
        )
        self.task = self.progress.add_task("Scanning...", total=total)

    def advance(self, amount: int) -> None:
        self.progress.advance(self.task, advance=amount)

    def __enter__(self) -> Self:
        self.progress.__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        self.progress.__exit__(*args)


class RichGGShieldUI(GGShieldUI):
    """
    Implementation of GGShieldUI using rich, for a more user-friendly terminal output.
    """

    def __init__(self):
        self.console = Console(file=sys.stderr)

    def create_scanner_ui(
        self,
        total: int,
        verbose: bool = False,
    ) -> ScannerUI:
        return RichProgressScannerUI(self, total, verbose)

    def create_message_only_scanner_ui(
        self,
        verbose: bool = False,
    ) -> ScannerUI:
        return RichMessageOnlyScannerUI(self, verbose)

    def create_progress(self, total: int) -> GGShieldProgress:
        return RichGGShieldProgress(self, total)

    def display_info(self, message: str) -> None:
        message = rich.markup.escape(message)
        self.console.print(message)

    def display_warning(self, message: str) -> None:
        message = rich.markup.escape(message)
        self.console.print(f"[yellow]Warning:[/] {message}")

    def display_error(self, message: str) -> None:
        message = rich.markup.escape(message)
        self.console.print(f"[red]Error:[/] {message}")



from abc import ABC, abstractmethod
from typing import Optional

import click

from ggshield.core.errors import ExitCode
from ggshield.verticals.sca.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
    SCAScanVulnerabilityCollection,
    SCAVulnerabilityCollectionType,
)


class SCAOutputHandler(ABC):
    verbose: bool
    exit_zero: bool
    output: Optional[str]

    def __init__(
        self,
        verbose: bool = False,
        exit_zero: bool = False,
        output: Optional[str] = None,
    ):
        self.verbose = verbose
        self.exit_zero = exit_zero
        self.output = output

    def process_scan_all_result(
        self, scan: SCAScanAllVulnerabilityCollection
    ) -> ExitCode:
        """Process a scan collection, write the report to :attr:`self.output`

        :param scan: The scan collection to process
        :return: The exit code
        """
        text = self._process_scan_all_impl(scan)
        return self._handle_process_scan_result(scan, text)

    def process_scan_diff_result(
        self, scan: SCAScanDiffVulnerabilityCollection
    ) -> ExitCode:
        """Process a scan collection, write the report to :attr:`self.output`

        :param scan: The scan collection to process
        :return: The exit code
        """
        text = self._process_scan_diff_impl(scan)
        return self._handle_process_scan_result(scan, text)

    @abstractmethod
    def _process_scan_all_impl(self, scan: SCAScanAllVulnerabilityCollection) -> str:
        """Implementation of scan processing,
        called by :meth:`OutputHandler.process_scan`

        Must return a string for the report.

        :param scan: The scan collection to process
        :return: The content
        """
        raise NotImplementedError()

    @abstractmethod
    def _process_scan_diff_impl(self, scan: SCAScanDiffVulnerabilityCollection) -> str:
        """Implementation of scan processing,
        called by :meth:`OutputHandler.process_scan`

        Must return a string for the report.

        :param scan: The scan collection to process
        :return: The content
        """
        raise NotImplementedError()

    def _get_exit_code(self, scan: SCAScanVulnerabilityCollection) -> ExitCode:
        if self.exit_zero:
            return ExitCode.SUCCESS
        if scan.result is None or scan.type == SCAVulnerabilityCollectionType.UNKNOWN:
            return ExitCode.UNEXPECTED_ERROR
        if scan.has_results:
            return ExitCode.SCAN_FOUND_PROBLEMS
        return ExitCode.SUCCESS

    def _handle_process_scan_result(
        self, scan: SCAScanVulnerabilityCollection, text: str
    ) -> ExitCode:
        if self.output:
            with open(self.output, "w+") as f:
                f.write(text)
        else:
            click.echo(text)
        return self._get_exit_code(scan)



"""
This module defines options which should be available on all commands, such as the
-v, --verbose option.

To use it:
- Add the `@add_common_options()` decorator after all the `click.option()` calls of the
  command function.
- Add a `**kwargs: Any` argument to the command function.

The `kwargs` argument is required because due to the way click works,
`add_common_options()` adds an argument for each option it defines.
"""
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar

import click

from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.debug_logs import setup_debug_logs
from ggshield.core.config.user_config import UserConfig


AnyFunction = Callable[..., Any]


# The argument of a Click option callback function
ArgT = TypeVar("ArgT")

# A Click option callback function
ClickCallback = Callable[
    [click.Context, click.Parameter, Optional[ArgT]], Optional[ArgT]
]


def get_config_from_context(ctx: click.Context) -> UserConfig:
    """Returns the UserConfig object stored in Click context"""
    return ContextObj.get(ctx).config.user_config


def create_ctx_callback(name: str) -> ClickCallback:
    """Helper function to define a Click option callback for simple cases where we only
    have to set a value on Click context object if the option is defined.
    """

    def callback(
        ctx: click.Context, param: click.Parameter, value: Optional[ArgT]
    ) -> Optional[ArgT]:
        if value is not None:
            setattr(ctx.obj, name, value)
        return value

    return callback


def create_config_callback(*option_names: str) -> ClickCallback:
    """Helper function to define a Click option callback for simple cases where we only
    have to set a configuration attribute if the option is defined.

    to reach UserConfig.foo, set option_names to ["foo"]
    to reach Userconfig.secret.bar, set option_names to ["secret", "bar"]
    """

    def callback(
        ctx: click.Context, param: click.Parameter, value: Optional[ArgT]
    ) -> Optional[ArgT]:
        if value is not None:
            obj = get_config_from_context(ctx)
            for name in option_names[:-1]:
                obj = getattr(obj, name)
            setattr(obj, option_names[-1], value)
        return value

    return callback


_verbose_option = click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=None,
    help="Verbose display mode.",
    callback=create_config_callback("verbose"),
)


def debug_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[bool]
) -> Optional[bool]:
    if value is not None:
        setup_debug_logs(filename=None)
    return value


# The --debug option is marked as "is_eager" so that we can setup logs as soon as
# possible. If we don't then log commands for the creation of the Config instance
# are ignored.
_debug_option = click.option(
    "--debug",
    is_flag=True,
    default=None,
    is_eager=True,
    help="Send log output to stderr. Equivalent to `--log-file -`.",
    callback=debug_callback,
)


def log_file_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[str]
) -> Optional[str]:
    if value is not None:
        setup_debug_logs(filename=None if value == "-" else value)
    return value


# The --log-file option is marked as "is_eager" so that we can setup logs as soon as
# possible. If we don't then log commands for the creation of the Config instance
# are ignored.
_log_file_option = click.option(
    "--log-file",
    metavar="FILE",
    is_eager=True,
    help="Send log output to FILE. Use '-' to redirect to stderr.",
    envvar="GITGUARDIAN_LOG_FILE",
    callback=log_file_callback,
)


_allow_self_signed_option = click.option(
    "--allow-self-signed",
    is_flag=True,
    default=None,
    help="Ignore ssl verification.",
    callback=create_config_callback("allow_self_signed"),
)

_check_for_updates = click.option(
    "--check-for-updates/--no-check-for-updates",
    is_flag=True,
    default=None,
    help="After executing commands, check if a new version of ggshield is available.",
    callback=create_ctx_callback("check_for_updates"),
)


exit_zero_option = click.option(
    "--exit-zero",
    is_flag=True,
    default=None,
    envvar="GITGUARDIAN_EXIT_ZERO",
    help=(
        "Always return a 0 (non-error) status code, even if incidents are found."
        " This option can also be set with the `GITGUARDIAN_EXIT_ZERO` environment"
        " variable."
    ),
    callback=create_config_callback("exit_zero"),
)


minimum_severity_option = click.option(
    "--minimum-severity",
    "minimum_severity",
    type=click.Choice(("LOW", "MEDIUM", "HIGH", "CRITICAL")),
    help="Minimum severity of the policies.",
)

ignore_path_option = click.option(
    "--ignore-path",
    "--ipa",
    "ignore_paths",
    default=None,
    multiple=True,
    help="Do not scan paths that match the specified glob-like patterns.",
    metavar="PATTERN",
)


def add_common_options() -> Callable[[AnyFunction], AnyFunction]:
    def decorator(cmd: AnyFunction) -> AnyFunction:
        _verbose_option(cmd)
        _debug_option(cmd)
        _log_file_option(cmd)
        _allow_self_signed_option(cmd)
        _check_for_updates(cmd)
        return cmd

    return decorator


json_option = click.option(
    "--json",
    "json_output",
    is_flag=True,
    default=None,
    help="Use JSON output.",
    callback=create_ctx_callback("use_json"),
)


directory_argument = click.argument(
    "directory",
    type=click.Path(exists=True, readable=True, path_type=Path, file_okay=False),
    required=False,
    # using a default value here makes the deprecated `iac scan` fail
)

all_option = click.option(
    "--all",
    "scan_all",
    is_flag=True,
    default=False,
    help="Reports all vulnerabilities in the final state.",
)

reference_option = click.option(
    "--ref",
    required=True,
    type=click.STRING,
    help="A Git reference, such as a commit ID, a reference relative to HEAD or a remote.",
    metavar="GIT_REF",
)
staged_option = click.option(
    "--staged",
    is_flag=True,
    help="Include staged changes in the scan.",
)



import os
from pathlib import Path
from typing import Any, Optional, Sequence

import click
from click import UsageError

from ggshield.cmd.sca.scan.sca_scan_utils import (
    create_output_handler,
    sca_scan_all,
    sca_scan_diff,
)
from ggshield.cmd.sca.scan.scan_common_options import (
    add_sca_scan_common_options,
    update_context,
)
from ggshield.cmd.utils.common_decorators import display_beta_warning, exception_wrapper
from ggshield.cmd.utils.common_options import all_option, directory_argument
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.errors import handle_exception
from ggshield.core.git_hooks.ci import get_current_and_previous_state_from_ci_env
from ggshield.core.git_hooks.ci.supported_ci import SupportedCI
from ggshield.core.scan.scan_mode import ScanMode
from ggshield.verticals.sca.collection.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
)


@click.command()
@add_sca_scan_common_options()
@click.pass_context
@directory_argument
@all_option
@display_beta_warning
@exception_wrapper
def scan_ci_cmd(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_paths: Sequence[str],
    directory: Optional[Path],
    scan_all: bool,
    **kwargs: Any,
) -> int:
    """
    Evaluates if a CI event introduces SCA vulnerabilities.

    Scanning a repository with this command will not trigger any incident on your dashboard.

    Only metadata such as call time, request size and scan mode is stored server-side.
    """
    if directory is None:
        directory = Path().resolve()

    # Adds client and required parameters to the context
    update_context(ctx, exit_zero, minimum_severity, ignore_paths)

    config = ContextObj.get(ctx).config
    try:
        if not (
            os.getenv("CI") or os.getenv("JENKINS_HOME") or os.getenv("BUILD_BUILDID")
        ):
            raise UsageError("`sca scan ci` should only be used in a CI environment.")

        output_handler = create_output_handler(ctx)
        if scan_all:
            result = sca_scan_all(ctx, directory, scan_mode=ScanMode.CI_ALL)
            scan = SCAScanAllVulnerabilityCollection(id=str(directory), result=result)
            return output_handler.process_scan_all_result(scan)

        current_commit, previous_commit = get_current_and_previous_state_from_ci_env(
            config.user_config.verbose
        )

        ci_mode = SupportedCI.from_ci_env()
        scan_mode = f"{ScanMode.CI_DIFF.value}/{ci_mode.value}"
        result = sca_scan_diff(
            ctx=ctx,
            directory=directory,
            previous_ref=previous_commit,
            current_ref=current_commit,
            scan_mode=scan_mode,
            ci_mode=ci_mode.name,
        )
        scan = SCAScanDiffVulnerabilityCollection(id=str(directory), result=result)
        return output_handler.process_scan_diff_result(scan)

    except Exception as error:
        return handle_exception(error, config.user_config.verbose)



import sys

from ggshield.core.ui.ggshield_ui import GGShieldProgress, GGShieldUI
from ggshield.core.ui.plain_text.plain_text_scanner_ui import PlainTextScannerUI
from ggshield.core.ui.scanner_ui import ScannerUI


class PlainTextGGShieldProgress(GGShieldProgress):
    def advance(self, amount: int) -> None:
        pass


class PlainTextGGShieldUI(GGShieldUI):
    """
    Plain-text implementation of GGShieldUI. Suitable when output is not a TTY.
    """

    def create_scanner_ui(
        self,
        total: int,
        verbose: bool = False,
    ) -> ScannerUI:
        return PlainTextScannerUI()

    def create_message_only_scanner_ui(
        self,
        verbose: bool = False,
    ) -> ScannerUI:
        return PlainTextScannerUI()

    def create_progress(self, total: int) -> GGShieldProgress:
        return PlainTextGGShieldProgress(self)

    def display_info(self, message: str) -> None:
        print(message, file=sys.stderr)

    def display_warning(self, message: str) -> None:
        print(f"WARNING: {message}", file=sys.stderr)

    def display_error(self, message: str) -> None:
        print(f"ERROR: {message}", file=sys.stderr)



import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Iterable, List, Optional, Set

from ggshield.utils.files import is_filepath_excluded
from ggshield.utils.git_shell import Filemode

from .scannable import Scannable


# Used instead of the SHA in commit URLs for staged changes
STAGED_PREFIX = "staged"

# Used instead of the SHA in commit URLs for commit files created from a patch
PATCH_PREFIX = "patch"

# Command line arguments passed to `git show` and `git diff` to get parsable headers
HEADER_COMMON_ARGS = [
    "--raw",  # shows a header with the files touched by the commit
    "-z",  # separate file names in the raw header with \0
    "-m",  # split multi-parent (aka merge) commits into several one-parent commits
]

# Command line arguments passed to `git show` and `git diff` to get parsable patches
PATCH_COMMON_ARGS = [
    *HEADER_COMMON_ARGS,
    "--patch",  # force output of the diff (--raw disables it)
]

# This block is inserted before the patch by commands generating patches from the staging
# area, where there is no commit info yet
DIFF_EMPTY_COMMIT_INFO_BLOCK = """Author:   <>\nDate:  \n:"""

_RX_HEADER_FILE_LINE_SEPARATOR = re.compile("[\n\0]:", re.MULTILINE)


class PatchParseError(Exception):
    """
    Raised by parse_patch() if it fails to parse its patch.
    """

    pass


class CommitScannable(Scannable):
    """Represents a file inside a commit. The URL of a CommitScannable looks like
    this:

        commit://<sha>/<path>

    For staged commits and commits from patches, `<sha>` is replaced with STAGED_PREFIX
    and PATCH_PREFIX respectively.
    """

    def __init__(
        self,
        sha: Optional[str],
        path: Path,
        content: str,
        filemode: Filemode = Filemode.MODIFY,
    ) -> None:
        super().__init__(filemode)
        self._sha = sha
        self._path = path
        self._content = content
        self._utf8_encoded_size = None

    @property
    def url(self) -> str:
        return CommitScannable.create_url(self._sha, self.path)

    @property
    def filename(self) -> str:
        return self.url

    @property
    def path(self) -> Path:
        return self._path

    def is_longer_than(self, max_utf8_encoded_size: int) -> bool:
        if self._utf8_encoded_size is None:
            self._utf8_encoded_size = len(self._content.encode(errors="replace"))
        return self._utf8_encoded_size > max_utf8_encoded_size

    @property
    def content(self) -> str:
        return self._content

    @staticmethod
    def create_url(sha: Optional[str], path: Path) -> str:
        """Creates a Commit URL based on the sha and Path. It is exposed as a static
        method so that code working with not-yet-parsed commits can return an URL for
        them."""
        prefix = sha if sha else STAGED_PREFIX
        return f"commit://{prefix}/{path.as_posix()}"


@dataclass
class PatchFileInfo:
    """
    Stores information about a file modified by a patch
    """

    # old_path is None unless filemode is RENAME or COPY
    old_path: Optional[Path]
    path: Path
    mode: Filemode

    @staticmethod
    def from_string(line: str) -> "PatchFileInfo":
        """
        Parse a file line in the raw patch header, returns a PatchFileInfo

        See https://github.com/git/git/blob/master/Documentation/diff-format.txt for
        details on the format.
        """

        prefix, path, *rest = line.rstrip("\0").split("\0")

        if rest:
            # If the line has a new path, it's a rename
            old_path = Path(path)
            new_path = Path(rest[0])
        else:
            old_path = None
            new_path = Path(path)

        # for a non-merge commit, prefix is
        # :old_perm new_perm old_sha new_sha status_and_score
        #
        # for a 2 parent commit, prefix is
        # ::old_perm1 old_perm2 new_perm old_sha1 old_sha2 new_sha status_and_score
        #
        # We can ignore most of it, because we only care about the status.
        #
        # status_and_score is one or more status letters, followed by an optional
        # numerical score. We can ignore the score, but we need to check the status
        # letters.
        status = prefix.rsplit(" ", 1)[-1].rstrip("0123456789")

        # There is one status letter per commit parent. In the case of a non-merge
        # commit the situation is simple: there is only one letter.
        # In the case of a merge commit we must look at all letters: if one parent is
        # marked as D(eleted) and the other as M(odified) then we use MODIFY as filemode
        # because the end result contains modifications. To ensure this, the order of
        # the `if` below matters.

        if "M" in status:  # modify
            mode = Filemode.MODIFY
        elif "C" in status:  # copy
            mode = Filemode.NEW
        elif "A" in status:  # add
            mode = Filemode.NEW
        elif "T" in status:  # type change
            mode = Filemode.NEW
        elif "R" in status:  # rename
            mode = Filemode.RENAME
        elif "D" in status:  # delete
            mode = Filemode.DELETE
        else:
            raise ValueError(f"Can't parse header line {line}: unknown status {status}")

        return PatchFileInfo(old_path, new_path, mode)


@dataclass
class PatchHeader:
    """
    Semi-parsed information about a patch. `info` is not parsed because some functions
    do not need the details in it.
    """

    # Meta information: commit author, date, message
    info: str

    # Files added/modified/removed by the commit
    files: List[PatchFileInfo] = field(default_factory=list)

    @staticmethod
    def from_string(header: str) -> "PatchHeader":
        # First item returned by split() contains commit info and message, skip it
        info, *lines = _RX_HEADER_FILE_LINE_SEPARATOR.split(header)
        return PatchHeader(
            info,
            [PatchFileInfo.from_string(x) for x in lines],
        )


def parse_patch(
    sha: Optional[str], patch: str, exclusion_regexes: Optional[Set[re.Pattern]]
) -> Iterable[Scannable]:
    """
    Parse a patch generated with `git show` or `git diff` using PATCH_COMMON_ARGS.

    If the patch represents a merge commit, then `patch` actually contains multiple
    commits, one per parent, because we call `git show` with the `-m` option to force it
    to generate one single-parent commit per parent. This makes later code simpler and
    ensures we see *all* the changes.
    """
    if exclusion_regexes is None:
        exclusion_regexes = set()

    for commit in patch.split("\0commit "):
        tokens = commit.split("\0diff ", 1)
        if len(tokens) == 1:
            # No diff, carry on to next commit
            continue
        header_str, rest = tokens

        try:
            header = PatchHeader.from_string(header_str)

            diffs = re.split(r"^diff ", rest, flags=re.MULTILINE)
            for file_info, diff in zip(header.files, diffs):
                if is_filepath_excluded(file_info.path, exclusion_regexes):
                    continue

                # extract document from diff: we must skip diff extended headers
                # (lines like "old mode 100644", "--- a/foo", "+++ b/foo"...)
                try:
                    end_of_headers = diff.index("\n@@")
                except ValueError:
                    # No content
                    continue
                # +1 because we searched for the '\n'
                content = diff[end_of_headers + 1 :]

                yield CommitScannable(
                    sha, file_info.path, content, filemode=file_info.mode
                )
        except Exception as exc:
            if sha:
                msg = f"Could not parse patch (sha: {sha}): {exc}"
            else:
                msg = f"Could not parse patch: {exc}"
            raise PatchParseError(msg)



import itertools
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import Callable, Iterable, Iterator, List, Optional, Set

from click import UsageError
from pygitguardian import GGClient

from ggshield.core.cache import Cache
from ggshield.core.client import check_client_api_key
from ggshield.core.config import Config
from ggshield.core.constants import MAX_WORKERS
from ggshield.core.errors import ExitCode, QuotaLimitReachedError, handle_exception
from ggshield.core.scan import Commit, ScanContext
from ggshield.core.types import IgnoredMatch
from ggshield.core.ui.ggshield_ui import GGShieldUI
from ggshield.utils.git_shell import get_list_commit_SHA, is_git_dir
from ggshield.utils.os import cd

from .output import SecretOutputHandler
from .secret_scan_collection import Results, SecretScanCollection
from .secret_scanner import SecretScanner


# We add a maximal value to avoid silently consuming all threads on powerful machines
SCAN_THREADS = 4


def scan_repo_path(
    client: GGClient,
    cache: Cache,
    ui: GGShieldUI,
    output_handler: SecretOutputHandler,
    config: Config,
    scan_context: ScanContext,
    repo_path: Path,
) -> int:  # pragma: no cover
    try:
        if not is_git_dir(repo_path):
            raise UsageError(f"{repo_path} is not a git repository")

        with cd(repo_path):
            return scan_commit_range(
                client=client,
                cache=cache,
                ui=ui,
                commit_list=get_list_commit_SHA("--all"),
                output_handler=output_handler,
                exclusion_regexes=set(),
                matches_ignore=config.user_config.secret.ignored_matches,
                scan_context=scan_context,
                ignored_detectors=config.user_config.secret.ignored_detectors,
                verbose=config.user_config.verbose,
            )
    except Exception as error:
        return handle_exception(error, config.user_config.verbose)


def scan_commits_content(
    commits: List[Commit],
    client: GGClient,
    cache: Cache,
    ui: GGShieldUI,
    matches_ignore: Iterable[IgnoredMatch],
    scan_context: ScanContext,
    progress_callback: Callable[[int], None],
    commit_scanned_callback: Callable[[Commit], None],
    ignored_detectors: Optional[Set[str]] = None,
) -> SecretScanCollection:  # pragma: no cover
    try:
        commit_files = itertools.chain.from_iterable(c.get_files() for c in commits)

        scanner = SecretScanner(
            client=client,
            cache=cache,
            scan_context=scan_context,
            ignored_matches=matches_ignore,
            ignored_detectors=ignored_detectors,
            check_api_key=False,  # Key has been checked in `scan_commit_range()`
        )
        with ui.create_message_only_scanner_ui() as scanner_ui:
            results = scanner.scan(
                commit_files, scan_threads=SCAN_THREADS, scanner_ui=scanner_ui
            )
    except QuotaLimitReachedError:
        raise
    except Exception as exc:
        results = Results.from_exception(exc)
    finally:
        progress_callback(len(commits))
        for commit in commits:
            commit_scanned_callback(commit)

    result_for_urls = {result.file.url: result for result in results.results}
    scans = []
    for commit in commits:
        results_for_commit_files = [
            result_for_urls[u] for u in commit.urls if u in result_for_urls
        ]
        scans.append(
            SecretScanCollection(
                commit.sha or "unknown",
                type="commit",
                results=Results(
                    results=results_for_commit_files,
                    errors=results.errors,
                ),
                optional_header=commit.optional_header,
                extra_info={
                    "author": commit.info.author,
                    "email": commit.info.email,
                    "date": commit.info.date,
                },
            )
        )

    return SecretScanCollection(
        id=scan_context.command_id, type="commit-ranges", scans=scans
    )


def get_commits_by_batch(
    commits: Iterable[Commit],
    batch_max_size: int,
) -> Iterator[List[Commit]]:
    """
    Given a list of commit shas yield the commit files
    by biggest batches possible of length at most `batch_max_size`
    """
    current_count = 0
    batch = []
    for commit in commits:
        num_files = len(commit.urls)
        if current_count + num_files < batch_max_size:
            batch.append(commit)
            current_count += num_files
        else:
            # The first batch can remain empty if it has too many files
            if batch:
                yield batch
            current_count = num_files
            batch = [commit]
    # Send the last batch that remains
    yield batch


def scan_commit_range(
    client: GGClient,
    cache: Cache,
    ui: GGShieldUI,
    commit_list: List[str],
    output_handler: SecretOutputHandler,
    exclusion_regexes: Set[re.Pattern],
    matches_ignore: Iterable[IgnoredMatch],
    scan_context: ScanContext,
    ignored_detectors: Optional[Set[str]] = None,
    verbose: bool = False,
) -> ExitCode:
    """
    Scan every commit in a range.

    :param client: Public Scanning API client
    :param commit_list: List of commits sha to scan
    :param verbose: Display successful scan's message
    """
    check_client_api_key(client)
    max_documents = client.secret_scan_preferences.maximum_documents_per_scan

    with ui.create_progress(len(commit_list)) as progress:
        commits_batch = get_commits_by_batch(
            commits=(
                Commit.from_sha(sha, exclusion_regexes=exclusion_regexes)
                for sha in commit_list
            ),
            batch_max_size=max_documents,
        )
        scans: List[SecretScanCollection] = []

        def commit_scanned_callback(commit: Commit):
            if verbose:
                progress.ui.display_info(f"Scanned {commit.sha}")

        with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
            futures = []
            for commits in commits_batch:
                futures.append(
                    executor.submit(
                        scan_commits_content,
                        commits,
                        client,
                        cache,
                        ui,
                        matches_ignore,
                        scan_context,
                        progress.advance,
                        commit_scanned_callback,
                        ignored_detectors,
                    )
                )
                # Stop now if an exception has been raised by a future
                for future in futures:
                    exception = future.exception()
                    if exception is not None:
                        raise exception

            for future in as_completed(futures):
                scan_collection = future.result()
                for scan in scan_collection.scans_with_results:
                    if scan.results and scan.results.errors:
                        for error in scan.results.errors:
                            progress.ui.display_error(error.description)
                    scans.append(scan)

    return_code = output_handler.process_scan(
        SecretScanCollection(
            id=scan_context.command_id, type="commit-range", scans=scans
        )
    )
    return return_code



import re
from pathlib import Path
from typing import List, Optional, Set

import click
from pygitguardian.client import GGClient
from pygitguardian.models import Detail

from ggshield.core.errors import APIKeyCheckError, UnexpectedError
from ggshield.core.scan import Scannable
from ggshield.core.scan.file import get_files_from_paths
from ggshield.core.tar_utils import INDEX_REF
from ggshield.core.text_utils import display_info
from ggshield.utils.files import is_filepath_excluded
from ggshield.utils.git_shell import get_filepaths_from_ref, get_staged_filepaths


# List of filepaths to ignore for SCA scans

SCA_IGNORE_LIST = (
    "__pycache__",
    ".git",
    ".hg",
    ".svn",
    ".tox",
    ".venv",
    "site-packages",
    ".idea",
    "node_modules",
    ".mypy_cache",
    ".pytest_cache",
    ".hypothesis",
)


def get_all_files_from_sca_paths(
    path: Path,
    exclusion_regexes: Set[re.Pattern],
    verbose: bool,
    ignore_git: bool = False,
) -> List[str]:
    """
    Create a Files object from a path, recursively, ignoring non SCA files

    :param path: path to scan
    :param exclusion_regexes: list of regexes, used to exclude some filepaths
    :param verbose: Option that displays filepaths as they are scanned
    :param ignore_git: Ignore that the folder is a git repository. If False, only files tracked by git are scanned
    """
    paths = [
        x.path
        for x in get_files_from_paths(
            paths=[path],
            exclusion_regexes=exclusion_regexes,
            recursive=True,
            yes=True,
            display_binary_files=verbose,
            display_scanned_files=False,  # If True, this displays all files in the directory but we only want SCA files
            ignore_git=ignore_git,
        )
        if is_not_excluded_from_sca(x)
    ]

    return [str(x.relative_to(path)) for x in paths]


def is_not_excluded_from_sca(scannable: Scannable) -> bool:
    """
    Returns True if file is in an SCA accepted path, which means that none of
    the directories of the path appear in SCA_IGNORE_LIST
    """
    return not any(part in SCA_IGNORE_LIST for part in scannable.path.parts)


def sca_files_from_git_repo(
    directory: Path,
    ref: str,
    client: GGClient,
    exclusion_regexes: Optional[Set[re.Pattern]] = None,
    verbose: bool = False,
) -> Set[Path]:
    """Returns SCA files from the git repository at
    the given directory, for the given ref. Empty string denotes selection
    from staging area."""
    exclusion_regexes = exclusion_regexes if exclusion_regexes is not None else set()

    if ref == INDEX_REF:
        all_files = get_staged_filepaths(wd=str(directory))
    else:
        all_files = get_filepaths_from_ref(ref, wd=str(directory))

    sca_files_result = client.compute_sca_files(
        files=[
            str(path)
            for path in all_files
            if not is_filepath_excluded(path, exclusion_regexes)
        ]
    )
    if isinstance(sca_files_result, Detail):
        if sca_files_result.status_code == 401:
            raise APIKeyCheckError(client.base_uri, "Invalid API key.")
        elif sca_files_result.status_code == 501:
            raise UnexpectedError(sca_files_result.detail)
        raise UnexpectedError("Failed to select SCA files")

    sca_files = sca_files_result.sca_files
    if verbose:
        display_info(f"> Scanned files from {ref}:")
        for filename in sca_files:
            display_info(f"- {click.format_filename(filename)}")

    return set(map(Path, sca_files_result.sca_files))



from typing import Optional

from pygitguardian.iac_models import IaCDiffScanEntities, IaCDiffScanResult

from ggshield.verticals.iac.collection.filter_ignored import filter_unignored_files
from ggshield.verticals.iac.collection.iac_scan_collection import (
    CollectionType,
    IaCScanCollection,
)


class IaCDiffScanCollection(IaCScanCollection):
    type = CollectionType.DiffScan
    result: Optional[IaCDiffScanResult]

    @property
    def has_results(self) -> bool:
        return self.result is not None and bool(self.result.entities_with_incidents.new)

    def get_entities_without_ignored(self) -> Optional[IaCDiffScanEntities]:
        if self.result is None:
            return None

        return IaCDiffScanEntities(
            new=filter_unignored_files(self.result.entities_with_incidents.new),
            unchanged=filter_unignored_files(
                self.result.entities_with_incidents.unchanged
            ),
            deleted=filter_unignored_files(self.result.entities_with_incidents.deleted),
        )

    def get_result_without_ignored(self) -> Optional[IaCDiffScanResult]:
        entities_without_ignored = self.get_entities_without_ignored()
        if self.result is None or entities_without_ignored is None:
            return None

        result_dict = self.result.to_dict()
        del result_dict["entities_with_incidents"]
        return IaCDiffScanResult(
            **result_dict, entities_with_incidents=entities_without_ignored
        )



import re
from pathlib import Path, PurePosixPath
from typing import List, Set, Union

from ggshield.utils._binary_extensions import BINARY_EXTENSIONS
from ggshield.utils.git_shell import get_filepaths_from_ref, git_ls, is_git_dir


class UnexpectedDirectoryError(ValueError):
    """Raise when a directory is used where it is not excepted"""

    def __init__(self, path: Path):
        self.path = path


def is_filepath_excluded(
    filepath: Union[str, Path], exclusion_regexes: Set[re.Pattern]
) -> bool:
    filepath = Path(filepath)
    return any(r.search(str(PurePosixPath(filepath))) for r in exclusion_regexes)


def get_filepaths(
    paths: List[Path],
    exclusion_regexes: Set[re.Pattern],
    recursive: bool,
    ignore_git: bool,
    ignore_git_staged: bool = False,
) -> Set[Path]:
    """
    Retrieve the filepaths from the command.

    :param paths: List of file/dir paths from the command
    :param recursive: Recursive option
    :param ignore_git: Ignore that the folder is a git repository
    :raise: UnexpectedDirectoryError if directory is given without --recursive option
    """
    targets: Set[Path] = set()
    for path in paths:
        if path.is_file():
            targets.add(path)
        elif path.is_dir():
            if not recursive:
                raise UnexpectedDirectoryError(path)

            if not ignore_git and is_git_dir(path):
                target_filepaths = (
                    get_filepaths_from_ref("HEAD", wd=path)
                    if ignore_git_staged
                    else git_ls(path)
                )
                _targets = {path / x for x in target_filepaths}
            else:
                _targets = path.rglob(r"*")

            for file_path in _targets:
                if not is_filepath_excluded(file_path, exclusion_regexes):
                    targets.add(file_path)
    return targets


def is_path_binary(path: Union[str, Path]) -> bool:
    ext = Path(path).suffix
    # `[1:]` because `ext` starts with a "." but extensions in `BINARY_EXTENSIONS` do not
    return ext[1:] in BINARY_EXTENSIONS



from abc import ABC, abstractmethod
from typing import Optional

import click

from ggshield.core.errors import ExitCode
from ggshield.core.text_utils import display_warning
from ggshield.verticals.iac.collection.iac_diff_scan_collection import (
    IaCDiffScanCollection,
)
from ggshield.verticals.iac.collection.iac_path_scan_collection import (
    IaCPathScanCollection,
)
from ggshield.verticals.iac.collection.iac_scan_collection import (
    CollectionType,
    IaCScanCollection,
)


class IaCOutputHandler(ABC):
    verbose: bool = False
    output: Optional[str] = None

    def __init__(
        self,
        verbose: bool,
        output: Optional[str] = None,
    ):
        self.verbose = verbose
        self.output = output

    def process_scan(self, scan: IaCPathScanCollection) -> ExitCode:
        """Process a scan collection, write the report to :attr:`self.output`

        :param scan: The scan collection to process
        :return: The exit code
        """
        text = self._process_scan_impl(scan)
        return self._handle_process_scan_result(scan, text)

    def process_diff_scan(self, scan: IaCDiffScanCollection) -> ExitCode:
        """Process a diff scan collection, write the report to :attr:`self.output`

        :param scan: The scan collection to process
        :return: The exit code
        """
        text = self._process_diff_scan_impl(scan)
        return self._handle_process_scan_result(scan, text)

    def process_skip_scan(self) -> ExitCode:
        """Process the case where we skip the scan,
        write the report to :attr:`self.output`

        :return: The exit code
        """
        text = self._process_skip_scan_impl()
        return self._handle_process_skip_scan(text)

    def process_skip_diff_scan(self) -> ExitCode:
        """Process the case where we skip the scan,
        write the report to :attr:`self.output`

        :return: The exit code
        """
        text = self._process_skip_diff_scan_impl()
        return self._handle_process_skip_scan(text)

    @abstractmethod
    def _process_skip_scan_impl(self) -> str:
        """Implementation of displaying a skipped scan,
        called by :meth:`OutputHandler.process_skip_scan`

        Must return a string for the report.

        :return: The content
        """
        raise NotImplementedError()

    @abstractmethod
    def _process_skip_diff_scan_impl(self) -> str:
        """Implementation of displaying a skipped diff scan,
        called by :meth:`OutputHandler.process_skip_diff_scan`

        Must return a string for the report.

        :return: The content
        """
        raise NotImplementedError()

    @abstractmethod
    def _process_scan_impl(self, scan: IaCPathScanCollection) -> str:
        """Implementation of scan processing,
        called by :meth:`OutputHandler.process_scan`

        Must return a string for the report.

        :param scan: The scan collection to process
        :return: The content
        """
        raise NotImplementedError()

    @abstractmethod
    def _process_diff_scan_impl(self, scan: IaCDiffScanCollection) -> str:
        """Implementation of diff scan processing,
        called by :meth:`OutputHandler.process_diff_scan`

        Must return a string for the report.

        :param scan: The scan collection to process
        :return: The content
        """
        raise NotImplementedError()

    def _get_exit_code(self, scan: IaCScanCollection) -> ExitCode:
        if scan.result is None or scan.type == CollectionType.Unknown:
            return ExitCode.UNEXPECTED_ERROR
        if scan.has_results:
            return ExitCode.SCAN_FOUND_PROBLEMS
        return ExitCode.SUCCESS

    def _handle_process_scan_result(
        self, scan: IaCScanCollection, text: str
    ) -> ExitCode:
        source_found = scan.result is not None and scan.result.source_found
        if self.verbose and not source_found:
            display_warning(
                "ggshield cannot fetch incidents monitored by the platform on this repository"
            )

        if self.output:
            with open(self.output, "w+") as f:
                f.write(text)
        else:
            click.echo(text)
        return self._get_exit_code(scan)

    def _handle_process_skip_scan(self, text: str) -> ExitCode:
        if self.output:
            with open(self.output, "w+") as f:
                f.write(text)
        else:
            click.echo(text)
        return ExitCode.SUCCESS



from typing import List, NamedTuple

from pygitguardian.models import Match

from ggshield.core.lines import Line


class MatchIndices(NamedTuple):
    line_index_start: int
    line_index_end: int
    index_start: int
    index_end: int


def find_match_indices(match: Match, lines: List[Line], is_patch: bool) -> MatchIndices:
    """Utility function.

    Returns a MatchIndices instance where
     - line_index_{start,end} are the indices in the lines of the line objects
       containing the start and end of the match
     - index_{start,end} are the indices of the match in the line_{start,end} objects

    :param match: a Match where index_{start,end} are not None
    :param lines: List of content lines with indices (post_index and pre_index)
    :param is_patch: True if is patch from git, False if file

    :return: MatchIndices
    """
    index = 0
    line_index = 0
    len_line = len(lines[line_index].content) + 1 + int(is_patch)
    # Update line_index until we find the secret start
    assert match.index_start is not None
    while match.index_start >= index + len_line:
        index += len_line
        line_index += 1
        len_line = len(lines[line_index].content) + 1 + int(is_patch)

    line_index_start = line_index
    index_start = match.index_start - index - int(is_patch)

    # Update line_index until we find the secret end
    assert match.index_end is not None
    while match.index_end > index + len_line:
        index += len_line
        line_index += 1
        len_line = len(lines[line_index].content) + 1 + int(is_patch)

    line_index_end = line_index
    index_end = match.index_end - index - int(is_patch) + 1
    return MatchIndices(
        line_index_start,
        line_index_end,
        index_start,
        index_end,
    )



from enum import Enum


class ScanMode(Enum):
    REPO = "repo"
    PATH = "path"
    COMMIT_RANGE = "commit_range"
    PRE_COMMIT = "pre_commit"
    PRE_PUSH = "pre_push"
    PRE_RECEIVE = "pre_receive"
    CI = "ci"
    DOCKER = "docker"
    PYPI = "pypi"
    ARCHIVE = "archive"
    DIRECTORY = "directory"
    DIFF = "diff"
    DOCSET = "docset"
    # IAC/SCA scan modes
    DIRECTORY_ALL = "directory_all"
    DIRECTORY_DIFF = "directory_diff"
    PRE_COMMIT_ALL = "pre_commit_all"
    PRE_COMMIT_DIFF = "pre_commit_diff"
    PRE_PUSH_ALL = "pre_push_all"
    PRE_PUSH_DIFF = "pre_push_diff"
    PRE_RECEIVE_ALL = "pre_receive_all"
    PRE_RECEIVE_DIFF = "pre_receive_diff"
    CI_ALL = "ci_all"
    CI_DIFF = "ci_diff"



"""
This module defines options which should be available on all sca scan subcommands.

To use it:
- Add the `@add_common_sca_scan_options()` decorator after all the `click.option()`
    calls of the command function.
- Add a `**kwargs: Any` argument to the command function.

The `kwargs` argument is required due to the way click works,
`add_common_options()` adds an argument for each option it defines.
"""
from typing import Callable, Sequence

import click

from ggshield.cmd.utils.common_options import (
    AnyFunction,
    add_common_options,
    exit_zero_option,
    ignore_path_option,
    json_option,
    minimum_severity_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.filter import init_exclusion_regexes


def add_sca_scan_common_options() -> Callable[[AnyFunction], AnyFunction]:
    def decorator(cmd: AnyFunction) -> AnyFunction:
        add_common_options()(cmd)
        exit_zero_option(cmd)
        minimum_severity_option(cmd)
        ignore_path_option(cmd)
        json_option(cmd)
        return cmd

    return decorator


def update_context(
    ctx: click.Context,
    exit_zero: bool,
    minimum_severity: str,
    ignore_paths: Sequence[str],
) -> None:
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    ctx_obj.client = create_client_from_config(config, ctx_obj.ui)

    if ignore_paths is not None:
        config.user_config.sca.ignored_paths.update(ignore_paths)

    ctx_obj.exclusion_regexes = init_exclusion_regexes(
        config.user_config.sca.ignored_paths
    )

    if exit_zero is not None:
        config.user_config.exit_zero = exit_zero

    if minimum_severity is not None:
        config.user_config.sca.minimum_severity = minimum_severity



import re
from pathlib import Path
from typing import List, Set

from ggshield.core.scan.file import get_files_from_paths


IAC_EXTENSIONS = {
    ".json",
    ".yml",
    ".yaml",
    ".jinja",
    ".py.schema",
    ".jinja.schema",
    ".tf",
}
IAC_FILENAME_KEYWORDS = {"tfvars", "dockerfile"}


def get_iac_files_from_path(
    path: Path,
    exclusion_regexes: Set[re.Pattern],
    verbose: bool,
    ignore_git: bool = False,
    ignore_git_staged: bool = False,
) -> List[Path]:
    """
    Returns IaC file paths found recursively in a given directory.

    :param path: root directory
    :param exclusion_regexes: Patterns to exclude from the files
    :param verbose: Option that displays filepaths as they are scanned
    :param ignore_git: Ignore that the folder is a git repository. If False, only files added to git are scanned
    """
    return [
        x.path
        for x in get_files_from_paths(
            paths=[path],
            exclusion_regexes=exclusion_regexes,
            recursive=True,
            yes=True,
            display_binary_files=verbose,
            display_scanned_files=False,  # If True, this displays all files in the directory but we only want IaC files
            ignore_git=ignore_git,
            ignore_git_staged=ignore_git_staged,
        )
        if is_iac_file_path(x.path)
    ]


def is_iac_file_path(path: Path) -> bool:
    if any(ext in IAC_EXTENSIONS for ext in path.suffixes):
        return True
    if any(path.name.endswith(iac_ext) for iac_ext in IAC_EXTENSIONS):
        return True
    name = path.name.lower()
    return any(keyword in name for keyword in IAC_FILENAME_KEYWORDS)



from typing import Any, Dict, cast

from ggshield.verticals.sca.collection.collection import (
    SCAScanAllVulnerabilityCollection,
    SCAScanDiffVulnerabilityCollection,
)
from ggshield.verticals.sca.output.handler import SCAOutputHandler
from ggshield.verticals.sca.output.schemas import (
    SCAJSONScanAllOutputSchema,
    SCAJSONScanDiffOutputSchema,
)


class SCAJsonOutputHandler(SCAOutputHandler):
    def _process_scan_all_impl(self, scan: SCAScanAllVulnerabilityCollection) -> str:
        scan_dict = self.create_scan_all_dict(scan)
        schema = SCAJSONScanAllOutputSchema
        serialized_result = schema.load(scan_dict)
        text = schema.dumps(serialized_result)
        return cast(str, text)

    def _process_scan_diff_impl(self, scan: SCAScanDiffVulnerabilityCollection) -> str:
        scan_dict = (
            scan.result.to_dict()
            if scan.result is not None
            else {"scanned_files": [], "added_vulns": [], "removed_vulns": []}
        )
        schema = SCAJSONScanDiffOutputSchema
        serialized_result = schema.load(scan_dict)
        text = schema.dumps(serialized_result)
        return cast(str, text)

    @staticmethod
    def create_scan_all_dict(scan: SCAScanAllVulnerabilityCollection) -> Dict[str, Any]:
        if scan.result is None:
            return {"scanned_files": [], "found_package_vulns": [], "total_vulns": 0}
        scan_dict = scan.result.to_dict()

        scan_dict["total_vulns"] = 0
        for file in scan_dict["found_package_vulns"]:
            scan_dict["total_vulns"] += len(file["package_vulns"])

        return scan_dict



import logging
import re
from dataclasses import dataclass, field
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import marshmallow_dataclass
from marshmallow import ValidationError, post_load, pre_load
from pygitguardian.models import FromDictMixin

from ggshield.core.config.utils import (
    find_global_config_path,
    find_local_config_path,
    load_yaml_dict,
    remove_common_dict_items,
    replace_in_keys,
    save_yaml_dict,
    update_from_other_instance,
)
from ggshield.core.constants import DEFAULT_LOCAL_CONFIG_PATH
from ggshield.core.errors import ParseError, UnexpectedError, format_validation_error
from ggshield.core.text_utils import display_warning
from ggshield.core.types import FilteredConfig, IgnoredMatch
from ggshield.core.url_utils import api_to_dashboard_url


logger = logging.getLogger(__name__)
CURRENT_CONFIG_VERSION = 2

_IGNORE_KNOWN_SECRETS_KEY = "ignore-known-secrets"

GHSA_ID_PATTERN = re.compile("GHSA(-[a-zA-Z0-9]{4}){3}")
POLICY_ID_PATTERN = re.compile("GG_IAC_[0-9]{4}")


@marshmallow_dataclass.dataclass
class SecretConfig(FilteredConfig):
    """
    Holds all user-defined secret-specific settings
    """

    show_secrets: bool = False
    ignored_detectors: Set[str] = field(default_factory=set)
    ignored_matches: List[IgnoredMatch] = field(default_factory=list)
    ignored_paths: Set[str] = field(default_factory=set)
    ignore_known_secrets: bool = False

    def add_ignored_match(self, secret: IgnoredMatch) -> None:
        """
        Add secret to ignored_matches.
        """
        for match in self.ignored_matches:
            if match.match == secret.match:
                # take the opportunity to name the ignored match
                if not match.name:
                    match.name = secret.name
                return
        self.ignored_matches.append(secret)


def validate_policy_id(policy_id: str) -> bool:
    return bool(POLICY_ID_PATTERN.fullmatch(policy_id))


@marshmallow_dataclass.dataclass
class ConfigIgnoredElement(FilteredConfig):
    """
    Base class for a config element which can be ignored temporarily.
    Classes inheriting it must reject expired dates themselves.
    """

    comment: Optional[str]
    until: Optional[datetime]

    # Accept date yyyy-mm-dd instead of a full datetime
    @pre_load
    def parse_date(self, data: Dict[str, Any], **kwargs: Any):
        if not (isinstance(data, dict)) or data.get("until") is None:
            return data
        try:
            data["until"] = str(datetime.strptime(str(data["until"]), "%Y-%m-%d"))
            return data
        except ValueError:
            return data

    @post_load
    def datetime_to_utc(self, data: Dict[str, Any], **kwargs: Any):
        if data["until"] is not None:
            data["until"] = data["until"].astimezone(timezone.utc)
        return data


def remove_expired_elements(
    lst: List[ConfigIgnoredElement],
) -> List[ConfigIgnoredElement]:
    expired: List[ConfigIgnoredElement] = []
    now = datetime.now(tz=timezone.utc)
    for idx in range(len(lst) - 1, -1, -1):
        ignored = lst[idx]
        if ignored.until and ignored.until <= now:
            lst.pop(idx)
            expired.insert(0, ignored)

    return expired


def report_expired_elements(expired_lst: List[ConfigIgnoredElement]) -> None:
    for element in expired_lst:
        display_warning(
            f"{element} has an expired 'until' "
            f"date ({element.until}), please update your configuration file."
        )


@marshmallow_dataclass.dataclass
class IaCConfigIgnoredPath(ConfigIgnoredElement):
    path: str

    def __init__(
        self,
        path: str,
        comment: Optional[str] = None,
        until: Optional[datetime] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(comment, until, *args, **kwargs)
        self.path = path

    def __str__(self):
        return f"Path {self.path}"

    @pre_load
    def convert_paths(self, in_data: Union[str, Dict[str, Any]], **kwargs: Any):
        return {"path": in_data} if isinstance(in_data, str) else in_data


@marshmallow_dataclass.dataclass
class IaCConfigIgnoredPolicy(ConfigIgnoredElement):
    policy: str = field(metadata={"validate": validate_policy_id})

    def __init__(
        self,
        policy: str,
        comment: Optional[str] = None,
        until: Optional[datetime] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(comment, until, *args, **kwargs)
        self.policy = policy

    def __str__(self):
        return f"Policy {self.policy}"

    @pre_load
    def convert_policies(self, in_data: Union[str, Dict[str, Any]], **kwargs: Any):
        return {"policy": in_data} if isinstance(in_data, str) else in_data


@marshmallow_dataclass.dataclass
class IaCConfig(FilteredConfig):
    """
    Holds the iac config as defined .gitguardian.yaml files
    (local and global).
    """

    ignored_paths: List[IaCConfigIgnoredPath] = field(default_factory=list)
    ignored_policies: List[IaCConfigIgnoredPolicy] = field(default_factory=list)
    minimum_severity: str = "LOW"

    # If we hit any outdated ignore rules when loading the configuration file, we
    # want to keep them in order to display additional details but we do not want
    # to store them in ignored rules because they will be used to scan.
    # Instead, we store them in these new computed fields. However we must mark them
    # as excluded, otherwise they would be serialized if the user runs a command to
    # write in the configuration file, which would be odd.
    outdated_ignored_paths: List[IaCConfigIgnoredPath] = field(default_factory=list)
    outdated_ignored_policies: List[IaCConfigIgnoredPolicy] = field(
        default_factory=list
    )

    @post_load
    def validate_ignored_paths(self, data: Dict[str, Any], **kwargs: Any):
        expired_lst = remove_expired_elements(data["ignored_paths"])
        report_expired_elements(expired_lst)
        data["outdated_ignored_paths"] = expired_lst
        return data

    @post_load
    def validate_ignored_policies(self, data: Dict[str, Any], **kwargs: Any):
        expired_lst = remove_expired_elements(data["ignored_policies"])
        report_expired_elements(expired_lst)
        data["outdated_ignored_policies"] = expired_lst
        return data


def is_ghsa_valid(ghsa_id: str) -> bool:
    return bool(GHSA_ID_PATTERN.fullmatch(ghsa_id))


def validate_vuln_identifier(value: str):
    if not is_ghsa_valid(value):
        raise ValidationError(
            f"The given GHSA id '{value}' do not match the pattern '{GHSA_ID_PATTERN.pattern}'"
        )


@marshmallow_dataclass.dataclass
class SCAConfigIgnoredVulnerability(ConfigIgnoredElement):
    """
    A model of an ignored vulnerability for SCA. This allows to ignore all occurrences
    of a given vulnerability in a given dependency file.
    - identifier: identifier (currently: GHSA id) of the vulnerability to ignore
    - path: the path to the file in which ignore the vulnerability
    - comment: The ignored reason
    - until: A datetime until which the vulnerability is ignored
    """

    identifier: str = field(metadata={"validate": validate_vuln_identifier})
    path: str

    def __init__(
        self,
        identifier: str,
        path: str,
        comment: Optional[str] = None,
        until: Optional[datetime] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(comment, until, *args, **kwargs)
        self.identifier = identifier
        self.path = path

    def __str__(self):
        return f"Vulnerability {self.identifier}"


@marshmallow_dataclass.dataclass
class SCAConfig(FilteredConfig):
    """
    Holds the sca config as defined .gitguardian.yaml files
    (local and global).
    """

    ignored_paths: Set[str] = field(default_factory=set)
    minimum_severity: str = "LOW"
    ignored_vulnerabilities: List[SCAConfigIgnoredVulnerability] = field(
        default_factory=list
    )

    @post_load
    def validate_ignored_vulns(self, data: Dict[str, Any], **kwargs: Any):
        expired_lst = remove_expired_elements(data["ignored_vulnerabilities"])
        report_expired_elements(expired_lst)
        return data


@marshmallow_dataclass.dataclass
class UserConfig(FilteredConfig):
    """
    Holds all ggshield settings defined by the user in the .gitguardian.yaml files
    (local and global).
    """

    iac: IaCConfig = field(default_factory=IaCConfig)
    instance: Optional[str] = None
    exit_zero: bool = False
    verbose: bool = False
    allow_self_signed: bool = False
    max_commits_for_hook: int = 50
    sca: SCAConfig = field(default_factory=SCAConfig)
    secret: SecretConfig = field(default_factory=SecretConfig)
    debug: bool = False

    # If we hit any deprecated syntax when loading a configuration file, we do not
    # display them directly, otherwise the messages would also be shown when running
    # `ggshield config migrate`, which would be odd.
    # Instead, we store them in this list and a result_callback() function displays
    # them when we quit. When `config migrate` runs, it clears this list, so nothing
    # gets displayed.
    deprecation_messages: List[str] = field(default_factory=list)

    def save(self, config_path: Path) -> None:
        """
        Save config to config_path
        """
        dct = self.to_dict()
        default_dct = UserConfig.from_dict({}).to_dict()

        dct = remove_common_dict_items(dct, default_dct)

        dct["version"] = CURRENT_CONFIG_VERSION

        replace_in_keys(dct, old_char="_", new_char="-")
        save_yaml_dict(dct, config_path)

    @classmethod
    def load(cls, config_path: Optional[Path] = None) -> Tuple["UserConfig", Path]:
        """
        Load the various user configs files to create a UserConfig object:
        - global user configuration file (in the home)
        - local user configuration file (in the repository)

        Returns a UserConfig instance, and the path where updates should be saved
        """
        user_config = UserConfig()
        if config_path:
            logger.debug("Loading custom config from %s", config_path)
            user_config._update_from_file(config_path)
            return user_config, config_path

        global_config_path = find_global_config_path()
        if global_config_path:
            user_config._update_from_file(global_config_path)
            logger.debug("Loaded global config from %s", global_config_path)
        else:
            logger.debug("No global config")

        local_config_path = find_local_config_path()
        if local_config_path:
            user_config._update_from_file(local_config_path)
            config_path = local_config_path
            logger.debug("Loaded local config from %s", local_config_path)
        else:
            logger.debug("No local config")

        if config_path is None:
            config_path = Path(DEFAULT_LOCAL_CONFIG_PATH)
        return user_config, config_path

    def _update_from_file(self, config_path: Path) -> None:
        try:
            data = load_yaml_dict(config_path) or {"version": CURRENT_CONFIG_VERSION}
            config_version = data.pop("version", 1)
            if config_version == 2:
                _fix_ignore_known_secrets(data)
                obj = UserConfig.from_dict(data)
            elif config_version == 1:
                replace_in_keys(data, old_char="-", new_char="_")
                self.deprecation_messages.append(
                    f"{config_path} uses a deprecated configuration file format."
                    " Run `ggshield config migrate` to migrate it to the latest version."
                )
                obj = UserV1Config.load_v1(data)
            else:
                raise UnexpectedError(
                    f"Don't know how to load config version {config_version}"
                )
        except ValidationError as exc:
            message = format_validation_error(exc)
            raise ParseError(message) from exc
        except ValueError as exc:
            raise ParseError(str(exc)) from exc

        update_from_other_instance(self, obj)


UserConfig.SCHEMA = marshmallow_dataclass.class_schema(UserConfig)(
    exclude=(
        "deprecation_messages",
        "iac.outdated_ignored_paths",
        "iac.outdated_ignored_policies",
    )
)


def _fix_ignore_known_secrets(data: Dict[str, Any]) -> None:
    """Fix a mistake done when implementing ignore-known-secrets: since this is a secret
    specific key, it should have been stored in the "secret" mapping, not in the root
    one"""

    underscore_key = _IGNORE_KNOWN_SECRETS_KEY.replace("-", "_")

    for key in _IGNORE_KNOWN_SECRETS_KEY, underscore_key:
        value = data.pop(key, None)
        if value is not None:
            break
    else:
        # No key to fix
        return

    secret_dct = data.setdefault("secret", {})
    if _IGNORE_KNOWN_SECRETS_KEY in secret_dct or underscore_key in secret_dct:
        # Do not override if it's already there
        return
    secret_dct[_IGNORE_KNOWN_SECRETS_KEY] = value


@dataclass
class UserV1Config(FromDictMixin):
    """
    Can load a v1 .gitguardian.yaml file
    """

    instance: Optional[str] = None
    all_policies: bool = False
    exit_zero: bool = False
    matches_ignore: List[Dict[str, Optional[str]]] = field(default_factory=list)
    paths_ignore: Set[str] = field(default_factory=set)
    verbose: bool = False
    allow_self_signed: bool = False
    max_commits_for_hook: int = 50
    ignore_default_excludes: bool = False
    show_secrets: bool = False
    banlisted_detectors: Set[str] = field(default_factory=set)

    @staticmethod
    def load_v1(data: Dict[str, Any]) -> UserConfig:
        """
        Takes a dict representing a v1 .gitguardian.yaml and returns a v2 config object
        """
        # If data contains the old "api-url" key, turn it into an "instance" key,
        # but only if there is no "instance" key
        try:
            api_url = data.pop("api_url")
        except KeyError:
            pass
        else:
            if "instance" not in data:
                data["instance"] = api_to_dashboard_url(api_url, warn=True)

        UserV1Config.matches_ignore_to_dict(data)

        v1config = UserV1Config.from_dict(data)

        deprecation_messages = []

        if v1config.all_policies:
            deprecation_messages.append(
                "The `all_policies` option has been deprecated and is now ignored."
            )

        if v1config.ignore_default_excludes:
            deprecation_messages.append(
                "The `ignore_default_exclude` option has been deprecated and is now ignored."
            )

        ignored_matches = [
            IgnoredMatch.from_dict(secret) for secret in v1config.matches_ignore
        ]
        secret = SecretConfig(
            show_secrets=v1config.show_secrets,
            ignored_detectors=v1config.banlisted_detectors,
            ignored_matches=ignored_matches,
            ignored_paths=v1config.paths_ignore,
        )

        return UserConfig(
            instance=v1config.instance,
            exit_zero=v1config.exit_zero,
            verbose=v1config.verbose,
            allow_self_signed=v1config.allow_self_signed,
            max_commits_for_hook=v1config.max_commits_for_hook,
            secret=secret,
            deprecation_messages=deprecation_messages,
        )

    @staticmethod
    def matches_ignore_to_dict(data: Dict[str, Any]) -> None:
        """
        v1 config format allowed to use just a hash of the secret for matches_ignore
        field v2 does not. This function converts the hash-only matches.
        """
        matches_ignore = data.get("matches_ignore")
        if not matches_ignore:
            return

        for idx, match in enumerate(matches_ignore):
            if isinstance(match, str):
                matches_ignore[idx] = {"name": "", "match": match}


UserV1Config.SCHEMA = marshmallow_dataclass.class_schema(UserV1Config)()



from abc import ABC, abstractmethod
from typing import Any

from typing_extensions import Self

from .scanner_ui import ScannerUI


class GGShieldProgress(ABC):
    """
    A generic progress bar. Must be created using `GGShieldUI.create_progress()` and
    used as a context manager.
    """

    def __init__(self, ui: "GGShieldUI"):
        self.ui = ui

    @abstractmethod
    def advance(self, amount: int) -> None:
        """
        Move the progress bar by `amount` steps.
        """
        ...

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args: Any) -> None:
        pass


class GGShieldUI(ABC):
    """Represents GGShield "UI", all progress-feedback code should ultimately go through
    this class (this is not the case at the moment).
    """

    @abstractmethod
    def create_scanner_ui(
        self,
        total: int,
        verbose: bool = False,
    ) -> ScannerUI:
        """
        Creates a ScannerUI instance. This is used to show progress on scanning
        Scannables.
        """
        ...

    @abstractmethod
    def create_message_only_scanner_ui(
        self,
        verbose: bool = False,
    ) -> ScannerUI:
        """
        Creates a ScannerUI instance without a progress bar. This is used when the scan
        itself is part of a larger scan. For example when scanning a commit range, each
        commit gets a message-only ScannerUI. Progress of the commit range scan is
        represented by a progress bar created using `create_progress()`.
        """
        ...

    @abstractmethod
    def create_progress(self, total: int) -> GGShieldProgress:
        """
        Creates a generic progress bar, not tied not scannables.
        """
        ...

    @abstractmethod
    def display_info(self, message: str) -> None:
        """
        Display an information message. Can be called while a progress bar is visible
        without messing the display.
        """
        ...

    @abstractmethod
    def display_warning(self, message: str) -> None:
        """
        Display a warning message. Can be called while a progress bar is visible without
        messing the display.
        """
        ...

    @abstractmethod
    def display_error(self, message: str) -> None:
        """
        Display an error message. Can be called while a progress bar is visible without
        messing the display.
        """
        ...



#!/usr/bin/python3
import logging
import os
import sys
from pathlib import Path
from typing import Any, List, Optional

import click

from ggshield.cmd.auth import auth_group
from ggshield.cmd.config import config_group
from ggshield.cmd.hmsl import hmsl_group
from ggshield.cmd.honeytoken import honeytoken_group
from ggshield.cmd.iac import iac_group
from ggshield.cmd.install import install_cmd
from ggshield.cmd.quota import quota_cmd
from ggshield.cmd.sca import sca_group
from ggshield.cmd.secret import secret_group
from ggshield.cmd.secret.scan import scan_group
from ggshield.cmd.status import status_cmd
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.debug_logs import disable_logs, setup_debug_logs
from ggshield.core import check_updates
from ggshield.core.cache import Cache
from ggshield.core.config import Config
from ggshield.core.env_utils import load_dot_env
from ggshield.core.errors import ExitCode
from ggshield.core.text_utils import display_warning
from ggshield.core.ui.plain_text.plain_text_ggshield_ui import PlainTextGGShieldUI
from ggshield.core.ui.rich.rich_ggshield_ui import RichGGShieldUI
from ggshield.utils.click import RealPath
from ggshield.utils.os import getenv_bool


logger = logging.getLogger(__name__)


@scan_group.result_callback()
@iac_group.result_callback()
@click.pass_context
def exit_code(ctx: click.Context, exit_code: int, **kwargs: Any) -> int:
    """
    exit_code guarantees that the return value of a scan is 0
    when exit_zero is enabled
    """
    ctx_obj = ContextObj.get(ctx)
    if (
        exit_code == ExitCode.SCAN_FOUND_PROBLEMS
        and ctx_obj.config.user_config.exit_zero
    ):
        logger.debug("scan exit_code forced to 0")
        sys.exit(ExitCode.SUCCESS)

    logger.debug("scan exit_code=%d", exit_code)
    return exit_code


def config_path_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[Path]
) -> Optional[Path]:
    # The --config option is marked as "is_eager" to ensure it's called before all the
    # others. This makes it the right place to create the configuration object.
    if not ctx.obj:
        ctx.obj = ContextObj()
        ctx.obj.cache = Cache()
        if sys.stderr.isatty():
            ctx.obj.ui = RichGGShieldUI()
        else:
            ctx.obj.ui = PlainTextGGShieldUI()

    ctx.obj.config = Config(value)
    return value


@click.group(
    context_settings={"help_option_names": ["-h", "--help"]},
    commands={
        "auth": auth_group,
        "config": config_group,
        "secret": secret_group,
        "install": install_cmd,
        "quota": quota_cmd,
        "api-status": status_cmd,
        "iac": iac_group,
        "honeytoken": honeytoken_group,
        "sca": sca_group,
        "hmsl": hmsl_group,
    },
)
@click.option(
    "-c",
    "--config-path",
    type=RealPath(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
    is_eager=True,
    help="Set a custom config file. Ignores local and global config files.",
    callback=config_path_callback,
)
@add_common_options()
@click.version_option()
@click.pass_context
def cli(
    ctx: click.Context,
    **kwargs: Any,
) -> None:
    load_dot_env()

    config = ContextObj.get(ctx).config

    _set_color(ctx)

    if config.user_config.debug:
        # if `debug` is set in the configuration file, then setup logs now.
        setup_debug_logs(filename=None)


def _set_color(ctx: click.Context):
    """
    Helper function to override the default click default output color setting.
        If NO_COLOR is set, we disable color output (see https://no-color.org/).
    If we are in a CI environment, certain variables are set, and we enable colors for
    the logs.
    """
    ci_env_vars = [
        "CI",  # Often set to indicate a generic CI environment
        "GITLAB_CI",
        "GITHUB_ACTIONS",
        "TRAVIS",
        "JENKINS_HOME",
        "JENKINS_URL",
        "CIRCLECI",
        "BITBUCKET_COMMIT",
        "DRONE",
        "BUILD_BUILDID",  # Azure Pipelines
    ]

    if os.getenv("NO_COLOR"):
        ctx.color = False
    elif any(os.getenv(env) for env in ci_env_vars):
        ctx.color = True


def _display_deprecation_message(cfg: Config) -> None:
    for message in cfg.user_config.deprecation_messages:
        display_warning(message)


def _check_for_updates(check_for_updates: bool) -> None:
    # Check for PYTEST_CURRENT_TEST to ensure update check does not happen when running
    # tests: we don't want it to happen because on the CI the unit test-suite is run
    # with --disable-socket, which causes failure on any network access.
    if check_for_updates and "PYTEST_CURRENT_TEST" not in os.environ:
        latest_version = check_updates.check_for_updates()
        if latest_version:
            display_warning(
                f"A new version of ggshield (v{latest_version}) has been released "
                f"(https://github.com/GitGuardian/ggshield)."
            )


@cli.result_callback()
@click.pass_context
def before_exit(ctx: click.Context, exit_code: int, *args: Any, **kwargs: Any) -> None:
    """
    This function is launched as a final callback once subcommands have run.
    It executes some final functions and then terminates.
    The argument exit_code is the result of the previously executed click command.
    """
    ctx_obj = ContextObj.get(ctx)
    _display_deprecation_message(ctx_obj.config)
    _check_for_updates(ctx_obj.check_for_updates)
    sys.exit(exit_code)


def main(args: Optional[List[str]] = None) -> Any:
    """
    Wrapper around cli.main() to handle the GITGUARDIAN_CRASH_LOG variable.

    `args` is only used by unit-tests.
    """
    disable_logs()
    show_crash_log = getenv_bool("GITGUARDIAN_CRASH_LOG")
    return cli.main(args, prog_name="ggshield", standalone_mode=not show_crash_log)


if __name__ == "__main__":
    sys.exit(main())



import logging
import os
from pathlib import Path
from typing import Any, Optional

import click

from ggshield.core.config.auth_config import AuthConfig
from ggshield.core.config.user_config import UserConfig
from ggshield.core.config.utils import remove_url_trailing_slash
from ggshield.core.constants import DEFAULT_HMSL_URL, DEFAULT_INSTANCE_URL
from ggshield.core.url_utils import (
    api_to_dashboard_url,
    clean_url,
    dashboard_to_api_url,
)


logger = logging.getLogger(__name__)


class Config:
    """
    Top-level config class. Contains an instance of UserConfig and an instance of
    AuthConfig.
    """

    __slots__ = ["user_config", "auth_config", "_cmdline_instance_name", "_config_path"]

    user_config: UserConfig
    auth_config: AuthConfig

    # The instance name, if ggshield is invoked with `--instance`
    _cmdline_instance_name: Optional[str]

    _config_path: Path

    def __init__(self, config_path: Optional[Path] = None):
        self.user_config, self._config_path = UserConfig.load(config_path=config_path)
        self.auth_config = AuthConfig.load()
        self._cmdline_instance_name = None

    def save(self) -> None:
        self.user_config.save(self._config_path)
        self.auth_config.save()

    @property
    def config_path(self) -> Path:
        return self._config_path

    @property
    def instance_name(self) -> str:
        """
        The instance name (defaulting to URL) of the selected instance
        priority order is:
        - set from the command line (by setting cmdline_instance_name)
        - env var (in auth_config.current_instance)
        - in local user config (in user_config.dashboard_url)
        - in global user config (in user_config.dashboard_url)
        - the default instance
        """
        if self._cmdline_instance_name:
            return self._cmdline_instance_name

        try:
            url = os.environ["GITGUARDIAN_INSTANCE"]
            logger.debug("Using instance URL from $GITGUARDIAN_INSTANCE")
        except KeyError:
            pass
        else:
            return remove_url_trailing_slash(url)

        try:
            name = os.environ["GITGUARDIAN_API_URL"]
            logger.debug("Using API URL from $GITGUARDIAN_API_URL")
        except KeyError:
            pass
        else:
            return api_to_dashboard_url(name, warn=True)

        if self.user_config.instance:
            return self.user_config.instance

        return DEFAULT_INSTANCE_URL

    @property
    def cmdline_instance_name(self) -> Optional[str]:
        return self._cmdline_instance_name

    @cmdline_instance_name.setter
    def cmdline_instance_name(self, name: str) -> None:
        """
        Override the current instance name. To be called by commands supporting the
        `--instance` option.
        """
        parsed_url = clean_url(name)
        if not parsed_url.scheme or not parsed_url.netloc:
            raise click.BadParameter(
                "Please provide a valid URL.",
                param_hint="instance",
            )
        self._cmdline_instance_name = f"{parsed_url.scheme}://{parsed_url.netloc}"

    @property
    def api_url(self) -> str:
        """
        The API URL to use the API
        It's the API URL from the configured instance
        """
        # TODO change when instance_name can be a name instead of just a URL
        return dashboard_to_api_url(self.instance_name)

    @property
    def dashboard_url(self) -> str:
        """
        The dashboard URL to use the dashboard
        It's the dashboard URL from the configured instance
        """
        # TODO change when instance_name can be a name instead of just a URL
        return self.instance_name

    @property
    def api_key(self) -> str:
        """
        The API key to use
        priority order is
        - env var
        - the api key from the selected instance
        """
        try:
            key = os.environ["GITGUARDIAN_API_KEY"]
            logger.debug("Using API key from $GITGUARDIAN_API_KEY")
        except KeyError:
            key = self.auth_config.get_instance_token(self.instance_name)
        return key

    def add_ignored_match(self, *args: Any, **kwargs: Any) -> None:
        return self.user_config.secret.add_ignored_match(*args, **kwargs)

    @property
    def saas_api_url(self) -> str:
        """
        The GIM SaaS instance (used to get JWT tokens).
        It is not configurable, but can be overridden with the
        GITGUARDIAN_SAAS_URL environment variable for tests.
        """
        if "api" in self.hmsl_url:  # case https://api.hasmysecretleaked.[...]
            default_value = self.hmsl_url.replace("hasmysecretleaked", "gitguardian")
        else:  # case https://hasmysecretleaked.[...]
            default_value = self.hmsl_url.replace("hasmysecretleaked", "api")

        return os.environ.get(
            "GITGUARDIAN_SAAS_URL",
            default_value,
        )

    @property
    def saas_api_key(self) -> str:
        """
        The API key associated with the SaaS GIM instance.
        For testing purposes, it can be overridden with the
        GITGUARDIAN_SAAS_API_KEY environment variable.
        """
        try:
            key = os.environ["GITGUARDIAN_SAAS_API_KEY"]
            logger.debug("Using API key from $GITGUARDIAN_SAAS_API_KEY")
        except KeyError:
            pass
        else:
            return key
        logger.debug("Using API key for SaaS instance from config")
        return self.api_key

    # Properties for HasMySecretLeaked
    # We can't rely on the instance selected by the user,
    # as JWT creation is only available in SaaS.
    @property
    def hmsl_url(self) -> str:
        """
        The url of the HasMySecretLeaked service
        """
        try:
            url = os.environ["GITGUARDIAN_HMSL_URL"]
            logger.debug("Using HasMySecretLeaked URL from $GITGUARDIAN_HMSL_URL")
        except KeyError:
            url = DEFAULT_HMSL_URL
            logger.debug("Using default HasMySecretLeaked URL")
        return remove_url_trailing_slash(url)

    @property
    def hmsl_audience(self) -> str:
        """
        The audience of our JWT tokens for HasMySecretLeaked.
        """
        try:
            audience = os.environ["GITGUARDIAN_HMSL_AUDIENCE"]
            logger.debug("Using audience from $GITGUARDIAN_HMSL_AUDIENCE")
        except KeyError:
            audience = self.hmsl_url
            logger.debug("Using HasMySecretLeaked URL as audience")
        return audience



import re
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Set, Union

import click

from ggshield.utils.files import UnexpectedDirectoryError, get_filepaths, is_path_binary

from .scannable import Scannable


class File(Scannable):
    """Implementation of Scannable for files from the disk."""

    def __init__(self, path: Union[str, Path]):
        super().__init__()
        self._path = Path(path)
        self._content: Optional[str] = None
        self._utf8_encoded_size: Optional[int] = None

    @property
    def url(self) -> str:
        return f"file://{self._path.absolute().as_posix()}"

    @property
    def filename(self) -> str:
        return str(self._path)

    @property
    def path(self) -> Path:
        return self._path

    def is_longer_than(self, max_utf8_encoded_size: int) -> bool:
        if self._utf8_encoded_size is not None:
            # We already have the encoded size, easy
            return self._utf8_encoded_size > max_utf8_encoded_size

        with self.path.open("rb") as fp:
            (
                result,
                self._content,
                self._utf8_encoded_size,
            ) = Scannable._is_file_longer_than(fp, max_utf8_encoded_size)
        return result

    @property
    def content(self) -> str:
        if self._content is None:
            with self.path.open("rb") as f:
                self._content, self._utf8_encoded_size = Scannable._decode_bytes(
                    f.read()
                )
        return self._content


def get_files_from_paths(
    paths: List[Path],
    exclusion_regexes: Set[re.Pattern],
    recursive: bool,
    yes: bool,
    display_scanned_files: bool,
    display_binary_files: bool,
    ignore_git: bool = False,
    ignore_git_staged: bool = False,
) -> List[Scannable]:
    """
    Create a scan object from files content.

    :param paths: List of file/dir paths from the command
    :param recursive: Recursive option
    :param yes: Skip confirmation option
    :param display_scanned_files: In some parts of the code (e.g. SCA), we might want
    to display a processed list instead and set this to False
    :param display_binary_files: Display all ignored binary files
    :param ignore_git: Ignore that the folder is a git repository
    """
    try:
        filepaths = get_filepaths(
            paths,
            exclusion_regexes,
            recursive,
            ignore_git=ignore_git,
            ignore_git_staged=ignore_git_staged,
        )
    except UnexpectedDirectoryError as error:
        raise click.UsageError(
            f"{click.format_filename(error.path)} is a directory."
            " Use --recursive to scan directories."
        )

    files = list(generate_files_from_paths(filepaths, display_binary_files))

    if display_scanned_files:
        for f in files:
            click.echo(f"- {click.format_filename(f.filename)}", err=True)

    size = len(files)
    if size > 1 and not yes:
        click.confirm(
            f"{size} files will be scanned. Do you want to continue?",
            abort=True,
            err=True,
        )

    return files


def generate_files_from_paths(
    paths: Iterable[Path], display_binary_files: bool
) -> Iterator[Scannable]:
    """Loop on filepaths and return an iterator on scannable files."""
    for path in paths:
        if path.is_dir() or not path.exists():
            continue

        if is_path_binary(path):
            if display_binary_files:
                click.echo(
                    f"ignoring binary file: {path}",
                    err=True,
                )
            continue

        yield File(path)



import re
from enum import Enum, auto
from typing import Iterable, List, NamedTuple, Optional

from ggshield.core.text_utils import STYLE, format_line_count, format_text
from ggshield.utils.git_shell import Filemode


REGEX_PATCH_HEADER = re.compile(
    r"^(?P<line_content>@@ -(?P<pre_index>\d+),?\d* \+(?P<post_index>\d+),?\d* @@(?: .+)?)"  # noqa
)

# Prefix used before each line for a complete file
FILE_LINE_PREFIX = "{} | "

# Prefix used before each line for a patch
PATCH_LINE_PREFIX = "{} {} | "


class LineCategory(Enum):
    addition = auto()
    data = auto()
    deletion = auto()
    empty = auto()


class Line(NamedTuple):
    """
    Line object making easier to handle line
    by line display.

    - content: Content of the line
    - category: The line category [+|-| ] (addition, deletion, untouched)
    - pre_index: Line index (deletion for patches, line index for files)
    - post_index: Line index (addition for patches)
    """

    content: str
    category: Optional[LineCategory] = None
    pre_index: Optional[int] = None
    post_index: Optional[int] = None

    def build_line_count(self, padding: int, is_secret: bool = False) -> str:
        """Return the formatted line count."""
        line_count_style = (
            STYLE["line_count_secret"] if is_secret else STYLE["line_count"]
        )
        if self.category is not None and not isinstance(self.category, LineCategory):
            raise TypeError("line category invalid")

        # File
        if self.category == LineCategory.data:
            return FILE_LINE_PREFIX.format(
                format_text(
                    format_line_count(self.pre_index, padding), line_count_style
                )
            )

        # Patch
        pre_index = format_line_count(self.pre_index, padding)
        post_index = format_line_count(self.post_index, padding)

        if self.category == LineCategory.addition:
            pre_index = " " * padding

        elif self.category == LineCategory.deletion:
            post_index = " " * padding

        return PATCH_LINE_PREFIX.format(
            format_text(pre_index, line_count_style),
            format_text(post_index, line_count_style),
        )


def get_lines_from_content(
    content: str, filemode: Filemode, is_patch: bool
) -> List[Line]:
    """
    Return the secrets and the lines with line number.

    :param content: Content to scan
    :param filemode: Filemode of the content
    :param is_patch: Is the content a patch
    """

    # Patch
    if is_patch:
        return list(get_lines_from_patch(content, filemode))

    # File
    return list(get_lines_from_file(content))


def get_lines_from_file(content: str) -> Iterable[Line]:
    """Return the lines with line number from a file."""
    for line_count, line_content in enumerate(content.split("\n")):
        yield Line(
            content=line_content, category=LineCategory.data, pre_index=line_count + 1
        )


def get_lines_from_patch(content: str, filemode: Filemode) -> Iterable[Line]:
    """Return the lines with line number from a git patch."""
    content += "\n"
    pre_index = 0
    post_index = 0

    for line in content.split("\n"):
        line_type = line[:1]
        line_content = ""
        line_pre_index = None
        line_post_index = None
        category = None

        if line_type == " ":
            line_content = line[1:]
            pre_index += 1
            post_index += 1
            line_pre_index = pre_index
            line_post_index = post_index
        elif line_type == "@":
            m = REGEX_PATCH_HEADER.search(line)
            if m is None:
                continue
            pre_index = int(m.groupdict()["pre_index"])
            post_index = int(m.groupdict()["post_index"])
            line_content = m.groupdict()["line_content"][:-1]

            if filemode == Filemode.NEW or filemode == Filemode.DELETE:
                pre_index = 1
                post_index = 1

            if line_content:
                line_type = " "
                pre_index -= 1
                post_index -= 1
                line_pre_index = None
                line_post_index = None
                category = LineCategory.empty
        elif line_type == "+":
            post_index += 1
            line_post_index = post_index
            line_content = line[1:]
            category = LineCategory.addition
        elif line_type == "-":
            pre_index += 1
            line_pre_index = pre_index
            line_content = line[1:]
            category = LineCategory.deletion
        elif line_type == "\\":
            # This type of line shouldn't contain any secret; no need to set indices
            line_content = line[1:]

        if line_type and line_content is not None:
            yield Line(
                content=line_content,
                category=category,
                pre_index=line_pre_index,
                post_index=line_post_index,
            )


def get_padding(lines: List[Line]) -> int:
    """Return the number of digit of the maximum line number."""
    # value can be None
    return max(len(str(lines[-1].pre_index or 0)), len(str(lines[-1].post_index or 0)))


def get_offset(padding: int, is_patch: bool = False) -> int:
    """Return the offset due to the line prefix."""
    if is_patch:
        return len(PATCH_LINE_PREFIX.format("0" * padding, "0" * padding))

    return len(FILE_LINE_PREFIX.format("0" * padding))
